# Copyright (C) 2005, 2006, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Topological sorting routines."""


from . import errors
from . import graph as _mod_graph
from . import revision as _mod_revision

__all__ = ["topo_sort", "TopoSorter", "merge_sort", "MergeSorter"]


def topo_sort(graph):
    """Topological sort a graph.

    graph -- sequence of pairs of node->parents_list.

    The result is a list of node names, such that all parents come before their
    children.

    node identifiers can be any hashable object, and are typically strings.

    This function has the same purpose as the TopoSorter class, but uses a
    different algorithm to sort the graph. That means that while both return a
    list with parents before their child nodes, the exact ordering can be
    different.

    topo_sort is faster when the whole list is needed, while when iterating
    over a part of the list, TopoSorter.iter_topo_order should be used.
    """
    kg = _mod_graph.KnownGraph(dict(graph))
    return kg.topo_sort()


class TopoSorter:

    def __init__(self, graph):
        """Topological sorting of a graph.

        :param graph: sequence of pairs of node_name->parent_names_list.
                      i.e. [('C', ['B']), ('B', ['A']), ('A', [])]
                      For this input the output from the sort or
                      iter_topo_order routines will be:
                      'A', 'B', 'C'

        node identifiers can be any hashable object, and are typically strings.

        If you have a graph like [('a', ['b']), ('a', ['c'])] this will only use
        one of the two values for 'a'.

        The graph is sorted lazily: until you iterate or sort the input is
        not processed other than to create an internal representation.

        iteration or sorting may raise GraphCycleError if a cycle is present
        in the graph.
        """
        # store a dict of the graph.
        self._graph = dict(graph)

    def sorted(self):
        """Sort the graph and return as a list.

        After calling this the sorter is empty and you must create a new one.
        """
        return list(self.iter_topo_order())

# Useful if fiddling with this code.
# cross check
###        sorted_names = list(self.iter_topo_order())
# for index in range(len(sorted_names)):
###            rev = sorted_names[index]
# for left_index in range(index):
# if rev in self.original_graph[sorted_names[left_index]]:
# print "revision in parent list of earlier revision"
###                    import pdb;pdb.set_trace()

    def iter_topo_order(self):
        """Yield the nodes of the graph in a topological order.

        After finishing iteration the sorter is empty and you cannot continue
        iteration.
        """
        graph = self._graph
        visitable = set(graph)

        # this is a stack storing the depth first search into the graph.
        pending_node_stack = []
        # at each level of 'recursion' we have to check each parent. This
        # stack stores the parents we have not yet checked for the node at the
        # matching depth in pending_node_stack
        pending_parents_stack = []

        # this is a set of the completed nodes for fast checking whether a
        # parent in a node we are processing on the stack has already been
        # emitted and thus can be skipped.
        completed_node_names = set()

        while graph:
            # now pick a random node in the source graph, and transfer it to the
            # top of the depth first search stack of pending nodes.
            node_name, parents = graph.popitem()
            pending_node_stack.append(node_name)
            pending_parents_stack.append(list(parents))

            # loop until pending_node_stack is empty
            while pending_node_stack:
                parents_to_visit = pending_parents_stack[-1]
                # if there are no parents left, the revision is done
                if not parents_to_visit:
                    # append the revision to the topo sorted list
                    # all the nodes parents have been added to the output,
                    # now we can add it to the output.
                    popped_node = pending_node_stack.pop()
                    pending_parents_stack.pop()
                    completed_node_names.add(popped_node)
                    yield popped_node
                else:
                    # recurse depth first into a single parent
                    next_node_name = parents_to_visit.pop()

                    if next_node_name in completed_node_names:
                        # parent was already completed by a child, skip it.
                        continue
                    if next_node_name not in visitable:
                        # parent is not a node in the original graph, skip it.
                        continue

                    # transfer it along with its parents from the source graph
                    # into the top of the current depth first search stack.
                    try:
                        parents = graph.pop(next_node_name)
                    except KeyError:
                        # if the next node is not in the source graph it has
                        # already been popped from it and placed into the
                        # current search stack (but not completed or we would
                        # have hit the continue 6 lines up).  this indicates a
                        # cycle.
                        raise errors.GraphCycleError(pending_node_stack)
                    pending_node_stack.append(next_node_name)
                    pending_parents_stack.append(list(parents))


def merge_sort(graph, branch_tip, mainline_revisions=None, generate_revno=False):
    """Topological sort a graph which groups merges.

    :param graph: sequence of pairs of node->parents_list.
    :param branch_tip: the tip of the branch to graph. Revisions not
                       reachable from branch_tip are not included in the
                       output.
    :param mainline_revisions: If not None this forces a mainline to be
                               used rather than synthesised from the graph.
                               This must be a valid path through some part
                               of the graph. If the mainline does not cover all
                               the revisions, output stops at the start of the
                               old revision listed in the mainline revisions
                               list.
                               The order for this parameter is oldest-first.
    :param generate_revno: Optional parameter controlling the generation of
        revision number sequences in the output. See the output description of
        the MergeSorter docstring for details.
    :result: See the MergeSorter docstring for details.

    Node identifiers can be any hashable object, and are typically strings.
    """
    return MergeSorter(graph, branch_tip, mainline_revisions,
                       generate_revno).sorted()


class MergeSorter:

    __slots__ = ['_node_name_stack',
                 '_node_merge_depth_stack',
                 '_pending_parents_stack',
                 '_first_child_stack',
                 '_left_subtree_pushed_stack',
                 '_generate_revno',
                 '_graph',
                 '_mainline_revisions',
                 '_stop_revision',
                 '_original_graph',
                 '_revnos',
                 '_revno_to_branch_count',
                 '_completed_node_names',
                 '_scheduled_nodes',
                 ]

    def __init__(self, graph, branch_tip, mainline_revisions=None,
                 generate_revno=False):
        """Merge-aware topological sorting of a graph.

        :param graph: sequence of pairs of node_name->parent_names_list.
                      i.e. [('C', ['B']), ('B', ['A']), ('A', [])]
                      For this input the output from the sort or
                      iter_topo_order routines will be:
                      'A', 'B', 'C'
        :param branch_tip: the tip of the branch to graph. Revisions not
                       reachable from branch_tip are not included in the
                       output.
        :param mainline_revisions: If not None this forces a mainline to be
                               used rather than synthesised from the graph.
                               This must be a valid path through some part
                               of the graph. If the mainline does not cover all
                               the revisions, output stops at the start of the
                               old revision listed in the mainline revisions
                               list.
                               The order for this parameter is oldest-first.
        :param generate_revno: Optional parameter controlling the generation of
            revision number sequences in the output. See the output description
            for more details.

        The result is a list sorted so that all parents come before
        their children. Each element of the list is a tuple containing:
        (sequence_number, node_name, merge_depth, end_of_merge)
         * sequence_number: The sequence of this row in the output. Useful for
           GUIs.
         * node_name: The node name: opaque text to the merge routine.
         * merge_depth: How many levels of merging deep this node has been
           found.
         * revno_sequence: When requested this field provides a sequence of
             revision numbers for all revisions. The format is:
             (REVNO, BRANCHNUM, BRANCHREVNO). BRANCHNUM is the number of the
             branch that the revno is on. From left to right the REVNO numbers
             are the sequence numbers within that branch of the revision.
             For instance, the graph {A:[], B:['A'], C:['A', 'B']} will get
             the following revno_sequences assigned: A:(1,), B:(1,1,1), C:(2,).
             This should be read as 'A is the first commit in the trunk',
             'B is the first commit on the first branch made from A', 'C is the
             second commit in the trunk'.
         * end_of_merge: When True the next node is part of a different merge.


        node identifiers can be any hashable object, and are typically strings.

        If you have a graph like [('a', ['b']), ('a', ['c'])] this will only use
        one of the two values for 'a'.

        The graph is sorted lazily: until you iterate or sort the input is
        not processed other than to create an internal representation.

        iteration or sorting may raise GraphCycleError if a cycle is present
        in the graph.

        Background information on the design:
        -------------------------------------
        definition: the end of any cluster or 'merge' occurs when:
            1 - the next revision has a lower merge depth than we do.
              i.e.
              A 0
              B  1
              C   2
              D  1
              E 0
              C, D are the ends of clusters, E might be but we need more data.
            2 - or the next revision at our merge depth is not our left most
              ancestor.
              This is required to handle multiple-merges in one commit.
              i.e.
              A 0    [F, B, E]
              B  1   [D, C]
              C   2  [D]
              D  1   [F]
              E  1   [F]
              F 0
              C is the end of a cluster due to rule 1.
              D is not the end of a cluster from rule 1, but is from rule 2: E
                is not its left most ancestor
              E is the end of a cluster due to rule 1
              F might be but we need more data.

        we show connecting lines to a parent when:
         - The parent is the start of a merge within this cluster.
           That is, the merge was not done to the mainline before this cluster
           was merged to the mainline.
           This can be detected thus:
            * The parent has a higher merge depth and is the next revision in
              the list.

          The next revision in the list constraint is needed for this case:
          A 0   [D, B]
          B  1  [C, F]   # we do not want to show a line to F which is depth 2
                           but not a merge
          C  1  [H]      # note that this is a long line to show back to the
                           ancestor - see the end of merge rules.
          D 0   [G, E]
          E  1  [G, F]
          F   2 [G]
          G  1  [H]
          H 0
         - Part of this merges 'branch':
          The parent has the same merge depth and is our left most parent and we
           are not the end of the cluster.
          A 0   [C, B] lines: [B, C]
          B  1  [E, C] lines: [C]
          C 0   [D]    lines: [D]
          D 0   [F, E] lines: [E, F]
          E  1  [F]    lines: [F]
          F 0
         - The end of this merge/cluster:
          we can ONLY have multiple parents at the end of a cluster if this
          branch was previously merged into the 'mainline'.
          - if we have one and only one parent, show it
            Note that this may be to a greater merge depth - for instance if
            this branch continued from a deeply nested branch to add something
            to it.
          - if we have more than one parent - show the second oldest (older ==
            further down the list) parent with
            an equal or lower merge depth
             XXXX revisit when awake. ddaa asks about the relevance of each one
             - maybe more than one parent is relevant
        """
        self._generate_revno = generate_revno
        # a dict of the graph.
        self._graph = dict(graph)
        # if there is an explicit mainline, alter the graph to match. This is
        # easier than checking at every merge whether we are on the mainline and
        # if so which path to take.
        if mainline_revisions is None:
            self._mainline_revisions = []
            self._stop_revision = None
        else:
            self._mainline_revisions = list(mainline_revisions)
            self._stop_revision = self._mainline_revisions[0]
        # skip the first revision, its what we reach and its parents are
        # therefore irrelevant
        for index, revision in enumerate(self._mainline_revisions[1:]):
            # NB: index 0 means self._mainline_revisions[1]
            # if the mainline matches the graph, nothing to do.
            parent = self._mainline_revisions[index]
            if parent is None:
                # end of mainline_revisions history
                continue
            graph_parent_ids = self._graph[revision]
            if not graph_parent_ids:
                # We ran into a ghost, skip over it, this is a workaround for
                # bug #243536, the _graph has had ghosts stripped, but the
                # mainline_revisions have not
                continue
            if graph_parent_ids[0] == parent:
                continue
            # remove it from its prior spot
            self._graph[revision].remove(parent)
            # insert it into the start of the mainline
            self._graph[revision].insert(0, parent)
        # we need to do a check late in the process to detect end-of-merges
        # which requires the parents to be accessible: its easier for now
        # to just keep the original graph around.
        self._original_graph = self._graph.copy()
        # we need to know the revision numbers of revisions to determine
        # the revision numbers of their descendants
        # this is a graph from node to [revno_tuple, first_child]
        # where first_child is True if no other children have seen this node
        # and revno_tuple is the tuple that was assigned to the node.
        # we dont know revnos to start with, so we start it seeded with
        # [None, True]
        self._revnos = {revision: [None, True]
                            for revision in self._graph}
        # Each mainline revision counts how many child branches have spawned from it.
        self._revno_to_branch_count = {}

        # this is a stack storing the depth first search into the graph.
        self._node_name_stack = []
        # at each level of recursion we need the merge depth this node is at:
        self._node_merge_depth_stack = []
        # at each level of 'recursion' we have to check each parent. This
        # stack stores the parents we have not yet checked for the node at the
        # matching depth in _node_name_stack
        self._pending_parents_stack = []
        # When we first look at a node we assign it a seqence number from its
        # leftmost parent.
        self._first_child_stack = []
        # this is a set of the nodes who have been completely analysed for fast
        # membership checking
        self._completed_node_names = set()
        # this is the scheduling of nodes list.
        # Nodes are scheduled
        # from the bottom left of the tree: in the tree
        # A 0  [D, B]
        # B  1 [C]
        # C  1 [D]
        # D 0  [F, E]
        # E  1 [F]
        # F 0
        # the scheduling order is: F, E, D, C, B, A
        # that is - 'left subtree, right subtree, node'
        # which would mean that when we schedule A we can emit the entire tree.
        self._scheduled_nodes = []
        # This records for each node when we have processed its left most
        # unmerged subtree. After this subtree is scheduled, all other subtrees
        # have their merge depth increased by one from this nodes merge depth.
        # it contains tuples - name, merge_depth
        self._left_subtree_pushed_stack = []

        # seed the search with the tip of the branch
        if (branch_tip is not None
            and branch_tip != _mod_revision.NULL_REVISION
                and branch_tip != (_mod_revision.NULL_REVISION,)):
            parents = self._graph.pop(branch_tip)
            self._push_node(branch_tip, 0, parents)

    def sorted(self):
        """Sort the graph and return as a list.

        After calling this the sorter is empty and you must create a new one.
        """
        return list(self.iter_topo_order())

    def iter_topo_order(self):
        """Yield the nodes of the graph in a topological order.

        After finishing iteration the sorter is empty and you cannot continue
        iteration.
        """
        # These are safe to offload to local variables, because they are used
        # as a stack and modified in place, never assigned to.
        node_name_stack = self._node_name_stack
        node_merge_depth_stack = self._node_merge_depth_stack
        pending_parents_stack = self._pending_parents_stack
        left_subtree_pushed_stack = self._left_subtree_pushed_stack
        completed_node_names = self._completed_node_names
        scheduled_nodes = self._scheduled_nodes

        graph_pop = self._graph.pop

        def push_node(node_name, merge_depth, parents,
                      node_name_stack_append=node_name_stack.append,
                      node_merge_depth_stack_append=node_merge_depth_stack.append,
                      left_subtree_pushed_stack_append=left_subtree_pushed_stack.append,
                      pending_parents_stack_append=pending_parents_stack.append,
                      first_child_stack_append=self._first_child_stack.append,
                      revnos=self._revnos,
                      ):
            """Add node_name to the pending node stack.

            Names in this stack will get emitted into the output as they are popped
            off the stack.

            This inlines a lot of self._variable.append functions as local
            variables.
            """
            node_name_stack_append(node_name)
            node_merge_depth_stack_append(merge_depth)
            left_subtree_pushed_stack_append(False)
            pending_parents_stack_append(list(parents))
            # as we push it, check if it is the first child
            parent_info = None
            if parents:
                # node has parents, assign from the left most parent.
                try:
                    parent_info = revnos[parents[0]]
                except KeyError:
                    # Left-hand parent is a ghost, consider it not to exist
                    pass
            if parent_info is not None:
                first_child = parent_info[1]
                parent_info[1] = False
            else:
                # We don't use the same algorithm here, but we need to keep the
                # stack in line
                first_child = None
            first_child_stack_append(first_child)

        def pop_node(node_name_stack_pop=node_name_stack.pop,
                     node_merge_depth_stack_pop=node_merge_depth_stack.pop,
                     first_child_stack_pop=self._first_child_stack.pop,
                     left_subtree_pushed_stack_pop=left_subtree_pushed_stack.pop,
                     pending_parents_stack_pop=pending_parents_stack.pop,
                     original_graph=self._original_graph,
                     revnos=self._revnos,
                     completed_node_names_add=self._completed_node_names.add,
                     scheduled_nodes_append=scheduled_nodes.append,
                     revno_to_branch_count=self._revno_to_branch_count,
                     ):
            """Pop the top node off the stack

            The node is appended to the sorted output.
            """
            # we are returning from the flattened call frame:
            # pop off the local variables
            node_name = node_name_stack_pop()
            merge_depth = node_merge_depth_stack_pop()
            first_child = first_child_stack_pop()
            # remove this node from the pending lists:
            left_subtree_pushed_stack_pop()
            pending_parents_stack_pop()

            parents = original_graph[node_name]
            parent_revno = None
            if parents:
                # node has parents, assign from the left most parent.
                try:
                    parent_revno = revnos[parents[0]][0]
                except KeyError:
                    # left-hand parent is a ghost, treat it as not existing
                    pass
            if parent_revno is not None:
                if not first_child:
                    # not the first child, make a new branch
                    base_revno = parent_revno[0]
                    branch_count = revno_to_branch_count.get(base_revno, 0)
                    branch_count += 1
                    revno_to_branch_count[base_revno] = branch_count
                    revno = (parent_revno[0], branch_count, 1)
                    # revno = (parent_revno[0], branch_count, parent_revno[-1]+1)
                else:
                    # as the first child, we just increase the final revision
                    # number
                    revno = parent_revno[:-1] + (parent_revno[-1] + 1,)
            else:
                # no parents, use the root sequence
                root_count = revno_to_branch_count.get(0, -1)
                root_count += 1
                if root_count:
                    revno = (0, root_count, 1)
                else:
                    revno = (1,)
                revno_to_branch_count[0] = root_count

            # store the revno for this node for future reference
            revnos[node_name][0] = revno
            completed_node_names_add(node_name)
            scheduled_nodes_append((node_name, merge_depth, revno))
            return node_name

        while node_name_stack:
            # loop until this call completes.
            parents_to_visit = pending_parents_stack[-1]
            # if all parents are done, the revision is done
            if not parents_to_visit:
                # append the revision to the topo sorted scheduled list:
                # all the nodes parents have been scheduled added, now
                # we can add it to the output.
                pop_node()
            else:
                while pending_parents_stack[-1]:
                    if not left_subtree_pushed_stack[-1]:
                        # recurse depth first into the primary parent
                        next_node_name = pending_parents_stack[-1].pop(0)
                        is_left_subtree = True
                        left_subtree_pushed_stack[-1] = True
                    else:
                        # place any merges in right-to-left order for scheduling
                        # which gives us left-to-right order after we reverse
                        # the scheduled queue. XXX: This has the effect of
                        # allocating common-new revisions to the right-most
                        # subtree rather than the left most, which will
                        # display nicely (you get smaller trees at the top
                        # of the combined merge).
                        next_node_name = pending_parents_stack[-1].pop()
                        is_left_subtree = False
                    if next_node_name in completed_node_names:
                        # this parent was completed by a child on the
                        # call stack. skip it.
                        continue
                    # otherwise transfer it from the source graph into the
                    # top of the current depth first search stack.
                    try:
                        parents = graph_pop(next_node_name)
                    except KeyError:
                        # if the next node is not in the source graph it has
                        # already been popped from it and placed into the
                        # current search stack (but not completed or we would
                        # have hit the continue 4 lines up.
                        # this indicates a cycle.
                        if next_node_name in self._original_graph:
                            raise errors.GraphCycleError(node_name_stack)
                        else:
                            # This is just a ghost parent, ignore it
                            continue
                    next_merge_depth = 0
                    if is_left_subtree:
                        # a new child branch from name_stack[-1]
                        next_merge_depth = 0
                    else:
                        next_merge_depth = 1
                    next_merge_depth = (
                        node_merge_depth_stack[-1] + next_merge_depth)
                    push_node(
                        next_node_name,
                        next_merge_depth,
                        parents)
                    # and do not continue processing parents until this 'call'
                    # has recursed.
                    break

        # We have scheduled the graph. Now deliver the ordered output:
        sequence_number = 0
        stop_revision = self._stop_revision
        generate_revno = self._generate_revno
        original_graph = self._original_graph

        while scheduled_nodes:
            node_name, merge_depth, revno = scheduled_nodes.pop()
            if node_name == stop_revision:
                return
            if not len(scheduled_nodes):
                # last revision is the end of a merge
                end_of_merge = True
            elif scheduled_nodes[-1][1] < merge_depth:
                # the next node is to our left
                end_of_merge = True
            elif (scheduled_nodes[-1][1] == merge_depth
                  and (scheduled_nodes[-1][0] not in
                       original_graph[node_name])):
                # the next node was part of a multiple-merge.
                end_of_merge = True
            else:
                end_of_merge = False
            if generate_revno:
                yield (sequence_number, node_name, merge_depth, revno, end_of_merge)
            else:
                yield (sequence_number, node_name, merge_depth, end_of_merge)
            sequence_number += 1

    def _push_node(self, node_name, merge_depth, parents):
        """Add node_name to the pending node stack.

        Names in this stack will get emitted into the output as they are popped
        off the stack.
        """
        self._node_name_stack.append(node_name)
        self._node_merge_depth_stack.append(merge_depth)
        self._left_subtree_pushed_stack.append(False)
        self._pending_parents_stack.append(list(parents))
        # as we push it, figure out if this is the first child
        parent_info = None
        if parents:
            # node has parents, assign from the left most parent.
            try:
                parent_info = self._revnos[parents[0]]
            except KeyError:
                # Left-hand parent is a ghost, consider it not to exist
                pass
        if parent_info is not None:
            first_child = parent_info[1]
            parent_info[1] = False
        else:
            # We don't use the same algorithm here, but we need to keep the
            # stack in line
            first_child = None
        self._first_child_stack.append(first_child)

    def _pop_node(self):
        """Pop the top node off the stack

        The node is appended to the sorted output.
        """
        # we are returning from the flattened call frame:
        # pop off the local variables
        node_name = self._node_name_stack.pop()
        merge_depth = self._node_merge_depth_stack.pop()
        first_child = self._first_child_stack.pop()
        # remove this node from the pending lists:
        self._left_subtree_pushed_stack.pop()
        self._pending_parents_stack.pop()

        parents = self._original_graph[node_name]
        parent_revno = None
        if parents:
            # node has parents, assign from the left most parent.
            try:
                parent_revno = self._revnos[parents[0]][0]
            except KeyError:
                # left-hand parent is a ghost, treat it as not existing
                pass
        if parent_revno is not None:
            if not first_child:
                # not the first child, make a new branch
                base_revno = parent_revno[0]
                branch_count = self._revno_to_branch_count.get(base_revno, 0)
                branch_count += 1
                self._revno_to_branch_count[base_revno] = branch_count
                revno = (parent_revno[0], branch_count, 1)
                # revno = (parent_revno[0], branch_count, parent_revno[-1]+1)
            else:
                # as the first child, we just increase the final revision
                # number
                revno = parent_revno[:-1] + (parent_revno[-1] + 1,)
        else:
            # no parents, use the root sequence
            root_count = self._revno_to_branch_count.get(0, 0)
            root_count = self._revno_to_branch_count.get(0, -1)
            root_count += 1
            if root_count:
                revno = (0, root_count, 1)
            else:
                revno = (1,)
            self._revno_to_branch_count[0] = root_count

        # store the revno for this node for future reference
        self._revnos[node_name][0] = revno
        self._completed_node_names.add(node_name)
        self._scheduled_nodes.append(
            (node_name, merge_depth, self._revnos[node_name][0]))
        return node_name



# Copyright (C) 2007-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Unit tests for the breezy.help module."""

import re
import textwrap

from .. import (builtins, commands, config, errors, help, help_topics, i18n,
                plugin, tests)
from .test_i18n import ZzzTranslations


class TestErrors(tests.TestCase):

    def test_no_help_topic(self):
        error = help.NoHelpTopic("topic")
        self.assertEqualDiff("No help could be found for 'topic'. "
                             "Please use 'brz help topics' to obtain a list of topics.",
                             str(error))


class TestCommandHelp(tests.TestCase):
    """Tests for help on commands."""

    def assertCmdHelp(self, expected, cmd):
        self.assertEqualDiff(textwrap.dedent(expected), cmd.get_help_text())

    def test_command_help_includes_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
            _see_also = ['foo', 'bar']
        self.assertCmdHelp('''\
Purpose: A sample command.
Usage:   brz WithSeeAlso

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

See also: bar, foo
''',
                           cmd_WithSeeAlso())

    def test_get_help_text(self):
        """Commands have a get_help_text method which returns their help."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command."""
        self.assertCmdHelp('''\
Purpose: A sample command.
Usage:   brz Demo

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

''',
                           cmd_Demo())
        cmd = cmd_Demo()
        helptext = cmd.get_help_text()
        self.assertStartsWith(helptext,
                              'Purpose: A sample command.\n'
                              'Usage:   brz Demo')
        self.assertEndsWith(helptext,
                            '  -v, --verbose  Display more information.\n\n')

    def test_command_with_additional_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
            _see_also = ['foo', 'bar']
        cmd = cmd_WithSeeAlso()
        helptext = cmd.get_help_text(['gam'])
        self.assertEndsWith(
            helptext,
            '  -h, --help     Show help message.\n'
            '  -q, --quiet    Only display errors and warnings.\n'
            '  --usage        Show usage message and options.\n'
            '  -v, --verbose  Display more information.\n'
            '\n'
            'See also: bar, foo, gam\n')

    def test_command_only_additional_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
        cmd = cmd_WithSeeAlso()
        helptext = cmd.get_help_text(['gam'])
        self.assertEndsWith(
            helptext,
            '  -v, --verbose  Display more information.\n'
            '\n'
            'See also: gam\n')

    def test_get_help_topic(self):
        """The help topic for a Command is its name()."""
        class cmd_foo_bar(commands.Command):
            __doc__ = """A sample command."""
        cmd = cmd_foo_bar()
        self.assertEqual(cmd.name(), cmd.get_help_topic())

    def test_formatted_help_text(self):
        """Help text should be plain text by default."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            :Examples:
                Example 1::

                    cmd arg1

                Example 2::

                    cmd arg2

                A code block follows.

                ::

                    brz Demo something
            """
        cmd = cmd_Demo()
        helptext = cmd.get_help_text()
        self.assertEqualDiff('''\
Purpose: A sample command.
Usage:   brz Demo

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

Examples:
    Example 1:

        cmd arg1

    Example 2:

        cmd arg2

    A code block follows.

        brz Demo something

''',
                             helptext)
        helptext = cmd.get_help_text(plain=False)
        self.assertEqualDiff('''\
:Purpose: A sample command.
:Usage:   brz Demo

:Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

:Examples:
    Example 1::

        cmd arg1

    Example 2::

        cmd arg2

    A code block follows.

    ::

        brz Demo something

''',
                             helptext)

    def test_concise_help_text(self):
        """Concise help text excludes the descriptive sections."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            Blah blah blah.

            :Examples:
                Example 1::

                    cmd arg1
            """
        cmd = cmd_Demo()
        helptext = cmd.get_help_text()
        self.assertEqualDiff('''\
Purpose: A sample command.
Usage:   brz Demo

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

Description:
  Blah blah blah.

Examples:
    Example 1:

        cmd arg1

''',
                             helptext)
        helptext = cmd.get_help_text(verbose=False)
        self.assertEqualDiff('''\
Purpose: A sample command.
Usage:   brz Demo

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

See brz help Demo for more details and examples.

''',
                             helptext)

    def test_help_custom_section_ordering(self):
        """Custom descriptive sections should remain in the order given."""
        class cmd_Demo(commands.Command):
            __doc__ = """\
A sample command.

Blah blah blah.

:Formats:
  Interesting stuff about formats.

:Examples:
  Example 1::

    cmd arg1

:Tips:
  Clever things to keep in mind.
"""
        cmd = cmd_Demo()
        helptext = cmd.get_help_text()
        self.assertEqualDiff('''\
Purpose: A sample command.
Usage:   brz Demo

Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

Description:
  Blah blah blah.

Formats:
  Interesting stuff about formats.

Examples:
  Example 1:

    cmd arg1

Tips:
  Clever things to keep in mind.

''',
                             helptext)

    def test_help_text_custom_usage(self):
        """Help text may contain a custom usage section."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            :Usage:
                cmd Demo [opts] args

                cmd Demo -h

            Blah blah blah.
            """
        cmd = cmd_Demo()
        helptext = cmd.get_help_text()
        self.assertEqualDiff('''\
Purpose: A sample command.
Usage:
    cmd Demo [opts] args

    cmd Demo -h


Options:
  -h, --help     Show help message.
  -q, --quiet    Only display errors and warnings.
  --usage        Show usage message and options.
  -v, --verbose  Display more information.

Description:
  Blah blah blah.

''',
                             helptext)


class ZzzTranslationsForDoc(ZzzTranslations):

    _section_pat = re.compile(':\\w+:\\n\\s+')
    _indent_pat = re.compile('\\s+')

    def zzz(self, s):
        m = self._section_pat.match(s)
        if m is None:
            m = self._indent_pat.match(s)
        if m:
            return '{}zz{{{{{}}}}}'.format(m.group(0), s[m.end():])
        return 'zz{{%s}}' % s


class TestCommandHelpI18n(tests.TestCase):
    """Tests for help on translated commands."""

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, '_translations', ZzzTranslationsForDoc())

    def assertCmdHelp(self, expected, cmd):
        self.assertEqualDiff(textwrap.dedent(expected), cmd.get_help_text())

    def test_command_help_includes_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
            _see_also = ['foo', 'bar']
        self.assertCmdHelp('''\
zz{{:Purpose: zz{{A sample command.}}
}}zz{{:Usage:   brz WithSeeAlso
}}
zz{{:Options:
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
zz{{:See also: bar, foo}}
''',
                           cmd_WithSeeAlso())

    def test_get_help_text(self):
        """Commands have a get_help_text method which returns their help."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command."""
        self.assertCmdHelp('''\
zz{{:Purpose: zz{{A sample command.}}
}}zz{{:Usage:   brz Demo
}}
zz{{:Options:
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
''',
                           cmd_Demo())

    def test_command_with_additional_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
            _see_also = ['foo', 'bar']
        cmd = cmd_WithSeeAlso()
        helptext = cmd.get_help_text(['gam'])
        self.assertEndsWith(
            helptext, '''\
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
zz{{:See also: bar, foo, gam}}
''')

    def test_command_only_additional_see_also(self):
        class cmd_WithSeeAlso(commands.Command):
            __doc__ = """A sample command."""
        cmd = cmd_WithSeeAlso()
        helptext = cmd.get_help_text(['gam'])
        self.assertEndsWith(
            helptext, '''\
zz{{:Options:
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
zz{{:See also: gam}}
''')

    def test_help_custom_section_ordering(self):
        """Custom descriptive sections should remain in the order given."""
        # The help formatter expect the class name to start with 'cmd_'
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            Blah blah blah.

            :Formats:
              Interesting stuff about formats.

            :Examples:
              Example 1::

                cmd arg1

            :Tips:
              Clever things to keep in mind.
            """
        self.assertCmdHelp('''\
zz{{:Purpose: zz{{A sample command.}}
}}zz{{:Usage:   brz Demo
}}
zz{{:Options:
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
Description:
  zz{{zz{{Blah blah blah.}}

}}:Formats:
  zz{{Interesting stuff about formats.}}

Examples:
  zz{{Example 1::}}

    zz{{cmd arg1}}

Tips:
  zz{{Clever things to keep in mind.}}

''',
                           cmd_Demo())

    def test_help_text_custom_usage(self):
        """Help text may contain a custom usage section."""
        class cmd_Demo(commands.Command):
            __doc__ = """A sample command.

            :Usage:
                cmd Demo [opts] args

                cmd Demo -h

            Blah blah blah.
            """
        self.assertCmdHelp('''\
zz{{:Purpose: zz{{A sample command.}}
}}zz{{:Usage:
    zz{{cmd Demo [opts] args}}

    zz{{cmd Demo -h}}

}}
zz{{:Options:
  -h, --help     zz{{Show help message.}}
  -q, --quiet    zz{{Only display errors and warnings.}}
  --usage        zz{{Show usage message and options.}}
  -v, --verbose  zz{{Display more information.}}
}}
Description:
  zz{{zz{{Blah blah blah.}}

}}
''',
                           cmd_Demo())


class TestHelp(tests.TestCase):

    def setUp(self):
        super().setUp()
        commands.install_bzr_command_hooks()


class TestRegisteredTopic(TestHelp):
    """Tests for the RegisteredTopic class."""

    def test_contruct(self):
        """Construction takes the help topic name for the registered item."""
        # validate our test
        self.assertTrue('basic' in help_topics.topic_registry)
        topic = help_topics.RegisteredTopic('basic')
        self.assertEqual('basic', topic.topic)

    def test_get_help_text(self):
        """RegisteredTopic returns the get_detail results for get_help_text."""
        topic = help_topics.RegisteredTopic('commands')
        self.assertEqual(help_topics.topic_registry.get_detail('commands'),
                         topic.get_help_text())

    def test_get_help_text_with_additional_see_also(self):
        topic = help_topics.RegisteredTopic('commands')
        self.assertEndsWith(
            topic.get_help_text(['foo', 'bar']),
            '\n'
            'See also: bar, foo\n')

    def test_get_help_text_loaded_from_file(self):
        # Pick a known topic stored in an external file
        topic = help_topics.RegisteredTopic('authentication')
        self.assertStartsWith(topic.get_help_text(),
                              'Authentication Settings\n'
                              '=======================\n'
                              '\n')

    def test_get_help_topic(self):
        """The help topic for RegisteredTopic is its topic from construction."""
        topic = help_topics.RegisteredTopic('foobar')
        self.assertEqual('foobar', topic.get_help_topic())
        topic = help_topics.RegisteredTopic('baz')
        self.assertEqual('baz', topic.get_help_topic())


class TestTopicIndex(TestHelp):
    """Tests for the HelpTopicIndex class."""

    def test_default_constructable(self):
        index = help_topics.HelpTopicIndex()

    def test_get_topics_None(self):
        """Searching for None returns the basic help topic."""
        index = help_topics.HelpTopicIndex()
        topics = index.get_topics(None)
        self.assertEqual(1, len(topics))
        self.assertIsInstance(topics[0], help_topics.RegisteredTopic)
        self.assertEqual('basic', topics[0].topic)

    def test_get_topics_topics(self):
        """Searching for a string returns the matching string."""
        index = help_topics.HelpTopicIndex()
        topics = index.get_topics('topics')
        self.assertEqual(1, len(topics))
        self.assertIsInstance(topics[0], help_topics.RegisteredTopic)
        self.assertEqual('topics', topics[0].topic)

    def test_get_topics_no_topic(self):
        """Searching for something not registered returns []."""
        index = help_topics.HelpTopicIndex()
        self.assertEqual([], index.get_topics('nothing by this name'))

    def test_prefix(self):
        """TopicIndex has a prefix of ''."""
        index = help_topics.HelpTopicIndex()
        self.assertEqual('', index.prefix)


class TestConfigOptionIndex(TestHelp):
    """Tests for the HelpCommandIndex class."""

    def setUp(self):
        super().setUp()
        self.index = help_topics.ConfigOptionHelpIndex()

    def test_get_topics_None(self):
        """Searching for None returns an empty list."""
        self.assertEqual([], self.index.get_topics(None))

    def test_get_topics_no_topic(self):
        self.assertEqual([], self.index.get_topics('nothing by this name'))

    def test_prefix(self):
        self.assertEqual('configuration/', self.index.prefix)

    def test_get_topic_with_prefix(self):
        topics = self.index.get_topics('configuration/default_format')
        self.assertLength(1, topics)
        opt = topics[0]
        self.assertIsInstance(opt, config.Option)
        self.assertEqual('default_format', opt.name)


class TestCommandIndex(TestHelp):
    """Tests for the HelpCommandIndex class."""

    def test_default_constructable(self):
        index = commands.HelpCommandIndex()

    def test_get_topics_None(self):
        """Searching for None returns an empty list."""
        index = commands.HelpCommandIndex()
        self.assertEqual([], index.get_topics(None))

    def test_get_topics_rocks(self):
        """Searching for 'rocks' returns the cmd_rocks command instance."""
        index = commands.HelpCommandIndex()
        topics = index.get_topics('rocks')
        self.assertEqual(1, len(topics))
        self.assertIsInstance(topics[0], builtins.cmd_rocks)

    def test_get_topics_no_topic(self):
        """Searching for something that is not a command returns []."""
        index = commands.HelpCommandIndex()
        self.assertEqual([], index.get_topics('nothing by this name'))

    def test_prefix(self):
        """CommandIndex has a prefix of 'commands/'."""
        index = commands.HelpCommandIndex()
        self.assertEqual('commands/', index.prefix)

    def test_get_topic_with_prefix(self):
        """Searching for commands/rocks returns the rocks command object."""
        index = commands.HelpCommandIndex()
        topics = index.get_topics('commands/rocks')
        self.assertEqual(1, len(topics))
        self.assertIsInstance(topics[0], builtins.cmd_rocks)


class TestHelpIndices(tests.TestCase):
    """Tests for the HelpIndices class."""

    def test_default_search_path(self):
        """The default search path should include internal indexs."""
        indices = help.HelpIndices()
        self.assertEqual(4, len(indices.search_path))
        # help topics should be searched in first.
        self.assertIsInstance(indices.search_path[0],
                              help_topics.HelpTopicIndex)
        # with commands being search second.
        self.assertIsInstance(indices.search_path[1],
                              commands.HelpCommandIndex)
        # plugins are a third index.
        self.assertIsInstance(indices.search_path[2],
                              plugin.PluginsHelpIndex)
        # config options are a fourth index
        self.assertIsInstance(indices.search_path[3],
                              help_topics.ConfigOptionHelpIndex)

    def test_search_for_unknown_topic_raises(self):
        """Searching for an unknown topic should raise NoHelpTopic."""
        indices = help.HelpIndices()
        indices.search_path = []
        error = self.assertRaises(help.NoHelpTopic, indices.search, 'foo')
        self.assertEqual('foo', error.topic)

    def test_search_calls_get_topic(self):
        """Searching should call get_topics in all indexes in order."""
        calls = []

        class RecordingIndex:
            def __init__(self, name):
                self.prefix = name

            def get_topics(self, topic):
                calls.append(('get_topics', self.prefix, topic))
                return ['something']
        index = help.HelpIndices()
        index.search_path = [RecordingIndex('1'), RecordingIndex('2')]
        # try with None
        index.search(None)
        self.assertEqual([
            ('get_topics', '1', None),
            ('get_topics', '2', None),
            ],
            calls)
        # and with a string
        del calls[:]
        index.search('bar')
        self.assertEqual([
            ('get_topics', '1', 'bar'),
            ('get_topics', '2', 'bar'),
            ],
            calls)

    def test_search_returns_index_and_results(self):
        """Searching should return help topics with their index"""
        class CannedIndex:
            def __init__(self, prefix, search_result):
                self.prefix = prefix
                self.result = search_result

            def get_topics(self, topic):
                return self.result
        index = help.HelpIndices()
        index_one = CannedIndex('1', ['a'])
        index_two = CannedIndex('2', ['b', 'c'])
        index.search_path = [index_one, index_two]
        self.assertEqual([(index_one, 'a'), (index_two, 'b'), (index_two, 'c')],
                         index.search(None))

    def test_search_checks_for_duplicate_prefixes(self):
        """Its an error when there are multiple indices with the same prefix."""
        indices = help.HelpIndices()
        indices.search_path = [help_topics.HelpTopicIndex(),
                               help_topics.HelpTopicIndex()]
        self.assertRaises(errors.DuplicateHelpPrefix, indices.search, None)



# Copyright (C) 2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for control directory formats."""

from breezy import errors
from breezy.tests.per_controldir import TestCaseWithControlDir


class TestControlDir(TestCaseWithControlDir):

    def test_get_format_description(self):
        self.assertIsInstance(self.bzrdir_format.get_format_description(),
                              str)

    def test_is_supported(self):
        self.assertIsInstance(self.bzrdir_format.is_supported(), bool)

    def test_upgrade_recommended(self):
        self.assertIsInstance(self.bzrdir_format.upgrade_recommended, bool)

    def test_supports_transport(self):
        self.assertIsInstance(
            self.bzrdir_format.supports_transport(self.get_transport()), bool)

    def test_check_support_status(self):
        if not self.bzrdir_format.is_supported():
            self.assertRaises(
                (errors.UnsupportedFormatError, errors.UnsupportedVcs),
                self.bzrdir_format.check_support_status, False)
        else:
            self.bzrdir_format.check_support_status(True)
            self.bzrdir_format.check_support_status(False)



# Copyright (C) 2005 Aaron Bentley, Canonical Ltd
# <aaron.bentley@utoronto.ca>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


from typing import Callable, Iterable, Iterator, List, Optional


class IterableFileBase:
    """Create a file-like object from any iterable"""

    def __init__(self, iterable: Iterable[bytes]) -> None:
        object.__init__(self)
        self._iter = iter(iterable)
        self._buffer = b""
        self.done = False

    def read_n(self, length: int) -> bytes:
        """
        >>> IterableFileBase([b'This ', b'is ', b'a ', b'test.']).read_n(8)
        b'This is '
        """
        def test_length(result):
            if len(result) >= length:
                return length
            else:
                return None
        return self._read(test_length)

    def read_to(self, sequence: bytes, length: Optional[int] = None) -> bytes:
        """
        >>> f = IterableFileBase([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.read_to(b'\\n')
        b'Th\\n'
        >>> f.read_to(b'\\n')
        b'is is \\n'
        """
        def test_contents(result):
            if length is not None:
                if len(result) >= length:
                    return length
            try:
                return result.index(sequence) + len(sequence)
            except ValueError:
                return None
        return self._read(test_contents)

    def _read(self, result_length: Callable[[bytes], Optional[int]]) -> bytes:
        """
        Read data until result satisfies the condition result_length.
        result_length is a callable that returns None until the condition
        is satisfied, and returns the length of the result to use when
        the condition is satisfied.  (i.e. it returns the length of the
        subset of the first condition match.)
        """
        result = self._buffer
        while result_length(result) is None:
            try:
                result += next(self._iter)
            except StopIteration:
                self.done = True
                self._buffer = b""
                return result
        output_length = result_length(result)
        self._buffer = result[output_length:]
        return result[:output_length]

    def read_all(self) -> bytes:
        """
        >>> IterableFileBase([b'This ', b'is ', b'a ', b'test.']).read_all()
        b'This is a test.'
        """
        def no_stop(result):
            return None
        return self._read(no_stop)

    def push_back(self, contents: bytes) -> None:
        """
        >>> f = IterableFileBase([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.read_to(b'\\n')
        b'Th\\n'
        >>> f.push_back(b"Sh")
        >>> f.read_all()
        b'Shis is \\na te\\nst.'
        """
        self._buffer = contents + self._buffer


class IterableFile:
    """This class supplies all File methods that can be implemented cheaply."""

    def __init__(self, iterable: Iterable[bytes]) -> None:
        object.__init__(self)
        self._file_base = IterableFileBase(iterable)
        self._iter = self._make_iterator()
        self._closed = False
        self.softspace = 0

    def _make_iterator(self) -> Iterator[bytes]:
        while not self._file_base.done:
            self._check_closed()
            result = self._file_base.read_to(b'\n')
            if result != b'':
                yield result

    def _check_closed(self):
        if self.closed:
            raise ValueError("File is closed.")

    def close(self) -> None:
        """
        >>> f = IterableFile([b'This ', b'is ', b'a ', b'test.'])
        >>> f.closed
        False
        >>> f.close()
        >>> f.closed
        True
        """
        self._file_base.done = True
        self._closed = True

    closed = property(lambda x: x._closed)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # If there was an error raised, prefer the original one
        try:
            self.close()
        except BaseException:
            if exc_type is None:
                raise
        return False

    def flush(self) -> None:
        """No-op for standard compliance.
        >>> f = IterableFile([])
        >>> f.close()
        >>> f.flush()
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        self._check_closed()

    def __next__(self) -> bytes:
        """Implementation of the iterator protocol's next()

        >>> f = IterableFile([b'This \\n', b'is ', b'a ', b'test.'])
        >>> next(f)
        b'This \\n'
        >>> f.close()
        >>> next(f)
        Traceback (most recent call last):
        ValueError: File is closed.
        >>> f = IterableFile([b'This \\n', b'is ', b'a ', b'test.\\n'])
        >>> next(f)
        b'This \\n'
        >>> next(f)
        b'is a test.\\n'
        >>> next(f)
        Traceback (most recent call last):
        StopIteration
        """
        self._check_closed()
        return next(self._iter)

    def __iter__(self) -> Iterator[bytes]:
        """
        >>> list(IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.']))
        [b'Th\\n', b'is is \\n', b'a te\\n', b'st.']
        >>> f = IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.close()
        >>> list(f)
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        return self

    def read(self, length: Optional[int] = None) -> bytes:
        """
        >>> IterableFile([b'This ', b'is ', b'a ', b'test.']).read()
        b'This is a test.'
        >>> f = IterableFile([b'This ', b'is ', b'a ', b'test.'])
        >>> f.read(10)
        b'This is a '
        >>> f = IterableFile([b'This ', b'is ', b'a ', b'test.'])
        >>> f.close()
        >>> f.read(10)
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        self._check_closed()
        if length is None:
            return self._file_base.read_all()
        else:
            return self._file_base.read_n(length)

    def read_to(self, sequence: bytes, size: Optional[int] = None) -> bytes:
        """
        Read characters until a sequence is found, with optional max size.
        The specified sequence, if found, will be included in the result

        >>> f = IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.read_to(b'i')
        b'Th\\ni'
        >>> f.read_to(b'i')
        b's i'
        >>> f.close()
        >>> f.read_to(b'i')
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        self._check_closed()
        return self._file_base.read_to(sequence, size)

    def readline(self, size: Optional[int] = None) -> bytes:
        """
        >>> f = IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.readline()
        b'Th\\n'
        >>> f.readline(4)
        b'is i'
        >>> f.close()
        >>> f.readline()
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        return self.read_to(b'\n', size)

    def readlines(self, sizehint=None) -> List[bytes]:
        """
        >>> f = IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.readlines()
        [b'Th\\n', b'is is \\n', b'a te\\n', b'st.']
        >>> f = IterableFile([b'Th\\nis ', b'is \\n', b'a ', b'te\\nst.'])
        >>> f.close()
        >>> f.readlines()
        Traceback (most recent call last):
        ValueError: File is closed.
        """
        lines: List[bytes] = []
        while True:
            line = self.readline()
            if line == b"":
                return lines
            if sizehint is None:
                lines.append(line)
            elif len(line) < sizehint:
                lines.append(line)
                sizehint -= len(line)
            else:
                self._file_base.push_back(line)
                return lines


if __name__ == "__main__":
    import doctest
    doctest.testmod()



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Test the exporter."""

import gzip
import os
import tempfile

from .... import tests
from ..exporter import (_get_output_stream, check_ref_format,
                        sanitize_ref_name_for_git)
from . import FastimportFeature


class TestOutputStream(tests.TestCase):

    _test_needs_features = [FastimportFeature]

    def test_get_output_stream_stdout(self):
        # - returns standard out
        self.assertIsNot(None, _get_output_stream("-"))

    def test_get_source_gz(self):
        fd, filename = tempfile.mkstemp(suffix=".gz")
        os.close(fd)
        with _get_output_stream(filename) as stream:
            stream.write(b"bla")
        # files ending in .gz are automatically decompressed.
        with gzip.GzipFile(filename) as f:
            self.assertEqual(b"bla", f.read())

    def test_get_source_file(self):
        # other files are opened as regular files.
        fd, filename = tempfile.mkstemp()
        os.close(fd)
        with _get_output_stream(filename) as stream:
            stream.write(b"foo")
        with open(filename) as f:
            self.assertEqual("foo", f.read())


# from dulwich.tests.test_repository:
class CheckRefFormatTests(tests.TestCase):
    """Tests for the check_ref_format function.

    These are the same tests as in the git test suite.
    """

    def test_valid(self):
        self.assertTrue(check_ref_format(b'heads/foo'))
        self.assertTrue(check_ref_format(b'foo/bar/baz'))
        self.assertTrue(check_ref_format(b'refs///heads/foo'))
        self.assertTrue(check_ref_format(b'foo./bar'))
        self.assertTrue(check_ref_format(b'heads/foo@bar'))
        self.assertTrue(check_ref_format(b'heads/fix.lock.error'))

    def test_invalid(self):
        self.assertFalse(check_ref_format(b'foo'))
        self.assertFalse(check_ref_format(b'foo/.bar'))
        self.assertFalse(check_ref_format(b'heads/foo/'))
        self.assertFalse(check_ref_format(b'heads/foo.'))
        self.assertFalse(check_ref_format(b'./foo'))
        self.assertFalse(check_ref_format(b'.refs/foo'))
        self.assertFalse(check_ref_format(b'heads/foo..bar'))
        self.assertFalse(check_ref_format(b'heads/foo?bar'))
        self.assertFalse(check_ref_format(b'heads/foo.lock'))
        self.assertFalse(check_ref_format(b'heads/v@{ation'))
        self.assertFalse(check_ref_format(b'heads/foo\\bar'))
        self.assertFalse(check_ref_format(b'heads/foo\bar'))
        self.assertFalse(check_ref_format(b'heads/foo bar'))
        self.assertFalse(check_ref_format(b'heads/foo\020bar'))
        self.assertFalse(check_ref_format(b'heads/foo\177bar'))


class CheckRefnameRewriting(tests.TestCase):
    """Tests for sanitize_ref_name_for_git function"""

    def test_passthrough_valid(self):
        self.assertEqual(sanitize_ref_name_for_git(b'heads/foo'), b'heads/foo')
        self.assertEqual(sanitize_ref_name_for_git(
            b'foo/bar/baz'), b'foo/bar/baz')
        self.assertEqual(sanitize_ref_name_for_git(
            b'refs///heads/foo'), b'refs///heads/foo')
        self.assertEqual(sanitize_ref_name_for_git(b'foo./bar'), b'foo./bar')
        self.assertEqual(sanitize_ref_name_for_git(
            b'heads/foo@bar'), b'heads/foo@bar')
        self.assertEqual(sanitize_ref_name_for_git(
            b'heads/fix.lock.error'), b'heads/fix.lock.error')

    def test_rewrite_invalid(self):
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'foo./bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo/')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo.')))
        self.assertTrue(check_ref_format(sanitize_ref_name_for_git(b'./foo')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'.refs/foo')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo..bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo?bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo.lock')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/v@{ation')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo\bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo\\bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo\020bar')))
        self.assertTrue(check_ref_format(
            sanitize_ref_name_for_git(b'heads/foo\177bar')))



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""ChunkWriter: write compressed data out with a fixed upper bound."""

import zlib
from typing import Callable, List, Optional, Tuple
from zlib import Z_FINISH, Z_SYNC_FLUSH


class ChunkWriter:
    """ChunkWriter allows writing of compressed data with a fixed size.

    If less data is supplied than fills a chunk, the chunk is padded with
    NULL bytes. If more data is supplied, then the writer packs as much
    in as it can, but never splits any item it was given.

    The algorithm for packing is open to improvement! Current it is:
     - write the bytes given
     - if the total seen bytes so far exceeds the chunk size, flush.

    :cvar _max_repack: To fit the maximum number of entries into a node, we
        will sometimes start over and compress the whole list to get tighter
        packing. We get diminishing returns after a while, so this limits the
        number of times we will try.
        The default is to try to avoid recompressing entirely, but setting this
        to something like 20 will give maximum compression.

    :cvar _max_zsync: Another tunable nob. If _max_repack is set to 0, then you
        can limit the number of times we will try to pack more data into a
        node. This allows us to do a single compression pass, rather than
        trying until we overflow, and then recompressing again.
    """
    #    In testing, some values for bzr.dev::
    #        repack  time  MB   max   full
    #         1       7.5  4.6  1140  0
    #         2       8.4  4.2  1036  1
    #         3       9.8  4.1  1012  278
    #         4      10.8  4.1  728   945
    #        20      11.1  4.1  0     1012
    #        repack = 0
    #        zsync   time  MB    repack  stop_for_z
    #         0       5.0  24.7  0       6270
    #         1       4.3  13.2  0       3342
    #         2       4.9   9.6  0       2414
    #         5       4.8   6.2  0       1549
    #         6       4.8   5.8  1       1435
    #         7       4.8   5.5  19      1337
    #         8       4.4   5.3  81      1220
    #        10       5.3   5.0  260     967
    #        11       5.3   4.9  366     839
    #        12       5.1   4.8  454     731
    #        15       5.8   4.7  704     450
    #        20       5.8   4.6  1133    7

    #    In testing, some values for mysql-unpacked::
    #                next_bytes estim
    #        repack  time  MB    full    stop_for_repack
    #         1            15.4  0       3913
    #         2      35.4  13.7  0       346
    #        20      46.7  13.4  3380    0
    #        repack=0
    #        zsync                       stop_for_z
    #         0      29.5 116.5  0       29782
    #         1      27.8  60.2  0       15356
    #         2      27.8  42.4  0       10822
    #         5      26.8  25.5  0       6491
    #         6      27.3  23.2  13      5896
    #         7      27.5  21.6  29      5451
    #         8      27.1  20.3  52      5108
    #        10      29.4  18.6  195     4526
    #        11      29.2  18.0  421     4143
    #        12      28.0  17.5  702     3738
    #        15      28.9  16.5  1223    2969
    #        20      29.6  15.7  2182    1810
    #        30      31.4  15.4  3891    23

    # Tuple of (num_repack_attempts, num_zsync_attempts)
    # num_zsync_attempts only has meaning if num_repack_attempts is 0.
    _repack_opts_for_speed = (0, 8)
    _repack_opts_for_size = (20, 0)

    def __init__(self, chunk_size: int, reserved: int = 0, optimize_for_size: bool = False) -> None:
        """Create a ChunkWriter to write chunk_size chunks.

        :param chunk_size: The total byte count to emit at the end of the
            chunk.
        :param reserved: How many bytes to allow for reserved data. reserved
            data space can only be written to via the write(...,
            reserved=True).
        """
        self.chunk_size = chunk_size
        self.compressor = zlib.compressobj()
        self.bytes_in: List[bytes] = []
        self.bytes_list: List[bytes] = []
        self.bytes_out_len = 0
        # bytes that have been seen, but not included in a flush to out yet
        self.unflushed_in_bytes = 0
        self.num_repack = 0
        self.num_zsync = 0
        self.unused_bytes: Optional[bytes] = None
        self.reserved_size = reserved
        # Default is to make building fast rather than compact
        self.set_optimize(for_size=optimize_for_size)

    def finish(self) -> Tuple[List[bytes], Optional[bytes], int]:
        """Finish the chunk.

        This returns the final compressed chunk, and either None, or the
        bytes that did not fit in the chunk.

        :return: (compressed_bytes, unused_bytes, num_nulls_needed)

            * compressed_bytes: a list of bytes that were output from the
              compressor. If the compressed length was not exactly chunk_size,
              the final string will be a string of all null bytes to pad this
              to chunk_size
            * unused_bytes: None, or the last bytes that were added, which we
              could not fit.
            * num_nulls_needed: How many nulls are padded at the end
        """
        self.bytes_in = []
        out = self.compressor.flush(Z_FINISH)
        self.bytes_list.append(out)
        self.bytes_out_len += len(out)

        if self.bytes_out_len > self.chunk_size:
            raise AssertionError('Somehow we ended up with too much'
                                 ' compressed data, %d > %d'
                                 % (self.bytes_out_len, self.chunk_size))
        nulls_needed = self.chunk_size - self.bytes_out_len
        if nulls_needed:
            self.bytes_list.append(b"\x00" * nulls_needed)
        return self.bytes_list, self.unused_bytes, nulls_needed

    def set_optimize(self, for_size: bool = True) -> None:
        """Change how we optimize our writes.

        :param for_size: If True, optimize for minimum space usage, otherwise
            optimize for fastest writing speed.
        :return: None
        """
        if for_size:
            opts = ChunkWriter._repack_opts_for_size
        else:
            opts = ChunkWriter._repack_opts_for_speed
        self._max_repack, self._max_zsync = opts

    def _recompress_all_bytes_in(self, extra_bytes: Optional[bytes] = None) -> Tuple[List[bytes], int, "zlib._Compress"]:
        """Recompress the current bytes_in, and optionally more.

        :param extra_bytes: Optional, if supplied we will add it with
            Z_SYNC_FLUSH
        :return: (bytes_out, bytes_out_len, alt_compressed)

            * bytes_out: is the compressed bytes returned from the compressor
            * bytes_out_len: the length of the compressed output
            * compressor: An object with everything packed in so far, and
              Z_SYNC_FLUSH called.
        """
        compressor = zlib.compressobj()
        bytes_out: List[bytes] = []
        append = bytes_out.append
        compress = compressor.compress
        for accepted_bytes in self.bytes_in:
            out = compress(accepted_bytes)
            if out:
                append(out)
        if extra_bytes:
            out = compress(extra_bytes)
            out += compressor.flush(Z_SYNC_FLUSH)
            append(out)
        bytes_out_len = sum(map(len, bytes_out))
        return bytes_out, bytes_out_len, compressor

    def write(self, bytes: bytes, reserved: bool = False) -> bool:
        """Write some bytes to the chunk.

        If the bytes fit, False is returned. Otherwise True is returned
        and the bytes have not been added to the chunk.

        :param bytes: The bytes to include
        :param reserved: If True, we can use the space reserved in the
            constructor.
        """
        if self.num_repack > self._max_repack and not reserved:
            self.unused_bytes = bytes
            return True
        if reserved:
            capacity = self.chunk_size
        else:
            capacity = self.chunk_size - self.reserved_size
        comp = self.compressor

        # Check to see if the currently unflushed bytes would fit with a bit of
        # room to spare, assuming no compression.
        next_unflushed = self.unflushed_in_bytes + len(bytes)
        remaining_capacity = capacity - self.bytes_out_len - 10
        if (next_unflushed < remaining_capacity):
            # looks like it will fit
            out = comp.compress(bytes)
            if out:
                self.bytes_list.append(out)
                self.bytes_out_len += len(out)
            self.bytes_in.append(bytes)
            self.unflushed_in_bytes += len(bytes)
        else:
            # This may or may not fit, try to add it with Z_SYNC_FLUSH
            # Note: It is tempting to do this as a look-ahead pass, and to
            #       'copy()' the compressor before flushing. However, it seems
            #       that Which means that it is the same thing as increasing
            #       repack, similar cost, same benefit. And this way we still
            #       have the 'repack' knob that can be adjusted, and not depend
            #       on a platform-specific 'copy()' function.
            self.num_zsync += 1
            if self._max_repack == 0 and self.num_zsync > self._max_zsync:
                self.num_repack += 1
                self.unused_bytes = bytes
                return True
            out = comp.compress(bytes)
            out += comp.flush(Z_SYNC_FLUSH)
            self.unflushed_in_bytes = 0
            if out:
                self.bytes_list.append(out)
                self.bytes_out_len += len(out)

            # We are a bit extra conservative, because it seems that you *can*
            # get better compression with Z_SYNC_FLUSH than a full compress. It
            # is probably very rare, but we were able to trigger it.
            if self.num_repack == 0:
                safety_margin = 100
            else:
                safety_margin = 10
            if self.bytes_out_len + safety_margin <= capacity:
                # It fit, so mark it added
                self.bytes_in.append(bytes)
            else:
                # We are over budget, try to squeeze this in without any
                # Z_SYNC_FLUSH calls
                self.num_repack += 1
                (bytes_out, this_len,
                 compressor) = self._recompress_all_bytes_in(bytes)
                if self.num_repack >= self._max_repack:
                    # When we get *to* _max_repack, bump over so that the
                    # earlier > _max_repack will be triggered.
                    self.num_repack += 1
                if this_len + 10 > capacity:
                    (bytes_out, this_len,
                     compressor) = self._recompress_all_bytes_in()
                    self.compressor = compressor
                    # Force us to not allow more data
                    self.num_repack = self._max_repack + 1
                    self.bytes_list = bytes_out
                    self.bytes_out_len = this_len
                    self.unused_bytes = bytes
                    return True
                else:
                    # This fits when we pack it tighter, so use the new packing
                    self.compressor = compressor
                    self.bytes_in.append(bytes)
                    self.bytes_list = bytes_out
                    self.bytes_out_len = this_len
        return False
